﻿#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 @file Cnoid_TurtleBot2_Ctrl.py
 @brief Controller for Choreonoid TurtleBot2 Plugin
 @date $Date$

 @author East Japan Accounting Center Co., Ltd

 (c)2018 The University of Aizu
 This software is released under the New BSD License.

"""
import os
import sys
import time
import datetime

# Import RTM module
import RTC
import OpenRTM_aist

import csv
from logging import getLogger, StreamHandler, DEBUG, INFO

sys.path.append('.')

logger = getLogger(__name__)
handler = StreamHandler()
handler.setLevel(INFO)
logger.setLevel(DEBUG)
logger.addHandler(handler)

# file mode
FILE_MODE_READ = 'r'
EXT_CSV = '.csv'

# Import Service implementation class
# <rtc-template block="service_impl">

# </rtc-template>

# Import Service stub modules
# <rtc-template block="consumer_import">
# </rtc-template>


# This module's spesification
# <rtc-template block="module_spec">
cnoid_turtlebot2_ctrl_spec = ["implementation_id", "Cnoid_TurtleBot2_Ctrl", 
         "type_name",         "Cnoid_TurtleBot2_Ctrl", 
         "description",       "Controller for Choreonoid TurtleBot2 Plugin", 
         "version",           "1.0.0", 
         "vendor",            "EAC", 
         "category",          "Choreonoid", 
         "activity_type",     "STATIC", 
         "max_instance",      "1", 
         "language",          "Python", 
         "lang_type",         "SCRIPT",
         "conf.default.log_file_name", "data.csv",

         "conf.__widget__.log_file_name", "text",

         "conf.__type__.log_file_name", "string",

         ""]
# </rtc-template>


class DataListener(OpenRTM_aist.ConnectorDataListenerT):
    def __init__(self, name):
        self._name = name

    def __del__(self):
        logger.info("dtor of %s", self._name)

    def __call__(self, info, cdrdata):
        logger.info("------------------------------")
        logger.info("Listener:       %s", self._name)
        logger.info("Profile::name:  %s", info.name)
        logger.info("Profile::id:    %s", info.id)
        logger.info("------------------------------")


class ConnListener(OpenRTM_aist.ConnectorListener):
    def __init__(self, name):
        self._name = name

    def __del__(self):
        logger.info("dtor of %s", self._name)

    def __call__(self, info):
        logger.info("------------------------------")
        logger.info("Listener:       %s", self._name)
        logger.info("Profile::name:  %s", info.name)
        logger.info("Profile::id:    %s", info.id)
        logger.info("------------------------------")


##
# @class Cnoid_TurtleBot2_Ctrl
# @brief Controller for Choreonoid TurtleBot2 Plugin
# 
# Overview:  Refer to the "機能仕様書_Cnoid_TurtleBot2_Ctrl_RTC.pdf" for details.
# 
# In / Out:  Refer to the "機能仕様書_Cnoid_TurtleBot2_Ctrl_RTC.pdf" for details.
# 
# Algorythm:  Refer to the "機能仕様書_Cnoid_TurtleBot2_Ctrl_RTC.pdf" for details.
# 
# 
class Cnoid_TurtleBot2_Ctrl(OpenRTM_aist.DataFlowComponentBase):
    
    ##
    # @brief constructor
    # @param manager Maneger Object
    # 
    def __init__(self, manager):
        OpenRTM_aist.DataFlowComponentBase.__init__(self, manager)

        self._d_out_odometry = OpenRTM_aist.instantiateDataType(RTC.TimedPose2D)
        """
        Coordinate values indicating location of the TurtleBot2
         - Type: TimedPose2D
         - Number: 1
         - Unit: [m,rad]
        """
        self._out_odometryOut = OpenRTM_aist.OutPort("out_odometry", self._d_out_odometry)

        # initialize of configuration-data.
        # <rtc-template block="init_conf_param">
        """
        Input file name
         - Name:  log_file_name
         - DefaultValue: data.csv
        """
        self._log_file_name = ['data.csv']

        # </rtc-template>

        self._file = None
        self._reader = None
        self._read_row = 0
        self._write_ok = 0

    ##
    # Data initialize
    #
    # The initialize action (on CREATED->ALIVE transition)
    # formaer rtc_init_entry() 
    # 
    # @return RTC::ReturnCode_t
    # 
    #
    def onInitialize(self):
        # Bind variables and configuration variable
        self.bindParameter("log_file_name", self._log_file_name, "data.csv")
        
        # Set InPort buffers
        
        # Set OutPort buffers
        self.addOutPort("out_odometry",self._out_odometryOut)
        
        # Set service provider to Ports
        
        # Set service consumers to Ports
        
        # Set CORBA Service Ports
        self._set_connector_listener(self._out_odometryOut)
        
        return RTC.RTC_OK
    
        ##
        # File close
        # 
        # The finalize action (on ALIVE->END transition)
        # formaer rtc_exiting_entry()
        # 
        # @return RTC::ReturnCode_t
    
        # 
    def onFinalize(self):
    
        self._close_robot_log()

        return RTC.RTC_OK

    #    ##
    #    #
    #    # The startup action when ExecutionContext startup
    #    # former rtc_starting_entry()
    #    # 
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onStartup(self, ec_id):
    #
    #    return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The shutdown action when ExecutionContext stop
    #    # former rtc_stopping_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onShutdown(self, ec_id):
    #
    #    return RTC.RTC_OK
    
        ##
        # CSV file open
        #
        # The activated action (Active state entry action)
        # former rtc_active_entry()
        #
        # @param ec_id target ExecutionContext Id
        # 
        # @return RTC::ReturnCode_t
        #
        #
    def onActivated(self, ec_id):
    
        rtn = False

        if self._is_open_robot_log():
            if self._is_csv_read():
                header = next(self._reader)  # Skip header
                self._read_row = 0
                self._write_ok = 0
                rtn = True

        if rtn:
            return RTC.RTC_OK
        else:
            return RTC.RTC_ERROR
    
        ##
        # File close
        #
        # The deactivated action (Active state exit action)
        # former rtc_active_exit()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onDeactivated(self, ec_id):
    
        self._close_robot_log()

        return RTC.RTC_OK

        ##
        # Read one record from the CSV file and write the data port. Execute
    #    "onExecute()" until EOF is detected.
        #
        # The execution action that is invoked periodically
        # former rtc_active_do()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onExecute(self, ec_id):
    
        try:
            row = next(self._reader)
            self._read_row += 1
            self._set(row)
            self._write()
        except StopIteration:
            self.deactivate(ec_id)
            logger.info('[EOF] read row count:   %d', self._read_row)
            logger.info('[EOF] write data count: %d', self._write_ok)

        return RTC.RTC_OK

    #    ##
    #    #
    #    # The aborting action when main logic error occurred.
    #    # former rtc_aborting_entry()
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onAborting(self, ec_id):
    #
    #    return RTC.RTC_OK
    
        ##
        # File close
        #
        # The error action in ERROR state
        # former rtc_error_do()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onError(self, ec_id):
    
        self._close_robot_log()

        return RTC.RTC_OK
    
        ##
        # File close
        #
        # The reset action that is invoked resetting
        # This is same but different the former rtc_init_entry()
        #
        # @param ec_id target ExecutionContext Id
        #
        # @return RTC::ReturnCode_t
        #
        #
    def onReset(self, ec_id):
    
        self._close_robot_log()

        return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The state update action that is invoked after onExecute() action
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #

    #    #
    #def onStateUpdate(self, ec_id):
    #
    #    return RTC.RTC_OK
    
    #    ##
    #    #
    #    # The action that is invoked when execution context's rate is changed
    #    # no corresponding operation exists in OpenRTm-aist-0.2.0
    #    #
    #    # @param ec_id target ExecutionContext Id
    #    #
    #    # @return RTC::ReturnCode_t
    #    #
    #    #
    #def onRateChanged(self, ec_id):
    #
    #    return RTC.RTC_OK
    
    def _set_connector_listener(self, dp):

        # dp.addConnectorDataListener(OpenRTM_aist.ConnectorDataListenerType.ON_BUFFER_WRITE,
        #                                         DataListener("ON_BUFFER_WRITE"))
        dp.addConnectorDataListener(OpenRTM_aist.ConnectorDataListenerType.ON_BUFFER_FULL,
                                    DataListener("ON_BUFFER_FULL"))
        dp.addConnectorDataListener(OpenRTM_aist.ConnectorDataListenerType.ON_BUFFER_WRITE_TIMEOUT,
                                    DataListener("ON_BUFFER_WRITE_TIMEOUT"))
        # dp.addConnectorDataListener(OpenRTM_aist.ConnectorDataListenerType.ON_BUFFER_OVERWRITE,
        #                                         DataListener("ON_BUFFER_OVERWRITE"))
        # dp.addConnectorDataListener(OpenRTM_aist.ConnectorDataListenerType.ON_BUFFER_READ,
        #                                         DataListener("ON_BUFFER_READ"))
        # dp.addConnectorDataListener(OpenRTM_aist.ConnectorDataListenerType.ON_SEND,
        #                                         DataListener("ON_SEND"))
        # dp.addConnectorDataListener(OpenRTM_aist.ConnectorDataListenerType.ON_RECEIVED,
        #                                         DataListener("ON_RECEIVED"))
        dp.addConnectorDataListener(OpenRTM_aist.ConnectorDataListenerType.ON_RECEIVER_FULL,
                                    DataListener("ON_RECEIVER_FULL"))
        dp.addConnectorDataListener(OpenRTM_aist.ConnectorDataListenerType.ON_RECEIVER_TIMEOUT,
                                    DataListener("ON_RECEIVER_TIMEOUT"))
        dp.addConnectorDataListener(OpenRTM_aist.ConnectorDataListenerType.ON_RECEIVER_ERROR,
                                    DataListener("ON_RECEIVER_ERROR"))

        dp.addConnectorListener(OpenRTM_aist.ConnectorListenerType.ON_CONNECT,
                                ConnListener("ON_CONNECT"))
        dp.addConnectorListener(OpenRTM_aist.ConnectorListenerType.ON_DISCONNECT,
                                ConnListener("ON_DISCONNECT"))

    def _is_open_robot_log(self):

        rtn = False

        try:
            filename = self._log_file_name[0]
            if filename != '' and os.path.splitext(filename)[1] == EXT_CSV:
                self._file = open(filename, FILE_MODE_READ)
                rtn = True
            else:
                logger.error('File does not exist !')

        except IOError as e:
            logger.error('I/O Error {0} {1}'.format(e.errno, e.strerror))

        return rtn

    def _is_csv_read(self):

        rtn = False

        if self._file is not None:
            self._reader = csv.DictReader(self._file)
            if self._reader is not None:
                rtn = True

        return rtn

    def _close_robot_log(self):

        if self._file is not None:
            self._file.close()
            self._file = None
            self._reader = None

    def _set(self, row):

        self._set_rtc_timestamp(row['datetime'])
        self._d_out_odometry.data.position.x = float(row['odometry_x'])
        self._d_out_odometry.data.position.y = float(row['odometry_y'])
        self._d_out_odometry.data.heading = float(row['odometry_theta'])

    def _set_rtc_timestamp(self, str_datetime):

        # convert datetime(yyyy-mm-dd HH:mm:ss.SSSSSS) to rtctime
        row = str_datetime.split()
        date = row[0]
        date = date.replace("-","")
        time = row[1]
        time = time.replace(":", "").replace(".", "")

        year = int(date[0:4])
        month = int(date[4:6])
        day = int(date[6:8])
        hr = int(time[0:2])
        min = int(time[2:4])
        sec = int(time[4:6])
        usec = int(time[6:12])
        rtctime_sec, rtctime_nsec = self._datetime_to_rtctime(
            year, month, day, hr, min, sec, usec)

        self._d_out_odometry.tm = RTC.Time(rtctime_sec, rtctime_nsec)

    def _datetime_to_rtctime(self, year, month, day, hr, min, sec, usec):

        dt = datetime.datetime(year, month, day, hr, min, sec, usec)
        rtctime_sec = int(time.mktime(dt.timetuple()))
        rtctime_nsec = usec * 1000

        return rtctime_sec, rtctime_nsec

    def _write(self):

        if self._out_odometryOut.write():
            self._write_ok += 1
        else:
            logger.error('Odometry write err row count:   %d', self._read_row)


def Cnoid_TurtleBot2_CtrlInit(manager):
    profile = OpenRTM_aist.Properties(defaults_str=cnoid_turtlebot2_ctrl_spec)
    manager.registerFactory(profile,
                            Cnoid_TurtleBot2_Ctrl,
                            OpenRTM_aist.Delete)

def MyModuleInit(manager):
    Cnoid_TurtleBot2_CtrlInit(manager)

    # Create a component
    comp = manager.createComponent("Cnoid_TurtleBot2_Ctrl")

def main():
    mgr = OpenRTM_aist.Manager.init(sys.argv)
    mgr.setModuleInitProc(MyModuleInit)
    mgr.activateManager()
    mgr.runManager()

if __name__ == "__main__":
    main()

